//---     
#include "..\WriteOptimizationResults.mq5"
//---   
#include "Enums.mqh"
#include "Errors.mqh"
#include "TradeSignals.mqh"
#include "TradeFunctions.mqh"
#include "Auxiliary.mqh"
//+------------------------------------------------------------------+
//|                                  |
//+------------------------------------------------------------------+
void GetTestStatistics(double &stat_array[])
  {
//---     
   double profit_factor=0,sharpe_ratio=0;
//---
   stat_array[0]=TesterStatistics(STAT_PROFIT);                //     
   stat_array[1]=TesterStatistics(STAT_DEALS);                 //   
//---
   profit_factor=TesterStatistics(STAT_PROFIT_FACTOR);         //    STAT_GROSS_PROFIT/STAT_GROSS_LOSS 
   stat_array[2]=(profit_factor==DBL_MAX) ? 0 : profit_factor; //   
//---
   stat_array[3]=TesterStatistics(STAT_EXPECTED_PAYOFF);       //   
   stat_array[4]=TesterStatistics(STAT_EQUITY_DDREL_PERCENT);  //     
   stat_array[5]=TesterStatistics(STAT_RECOVERY_FACTOR);       //     STAT_PROFIT/STAT_BALANCE_DD
//---
   sharpe_ratio=TesterStatistics(STAT_SHARPE_RATIO);           //   -     ()
   stat_array[6]=(sharpe_ratio==DBL_MAX) ? 0 : sharpe_ratio;   //   
  }
//+------------------------------------------------------------------+
//|                     |
//+------------------------------------------------------------------+
void CreateOptimizationReport()
  {
   static int passes_count=0;                //  
   int        parameters_count=0;            //   
   int        optimized_parameters_count=0;  //   
   string     string_to_write="";            //   
   bool       include_criteria_list=false;   //     -
   int        equality_sign_index=0;         //   '='  
   string     name            ="";           //  / 
   ulong      pass            =0;            //    ,    
   long       id              =0;            //  id 
   double     value           =0.0;          //    
   string     parameters_list[];             //     "parameterN=valueN"
   string     parameter_names[];             //   
   string     parameter_values[];            //   
//---   
   passes_count++;
//---     
   FrameNext(pass,name,id,value,stat_values);
//---   ,  , - 
   FrameInputs(pass,parameters_list,parameters_count);
//---       (   )
//        ,     
   for(int i=0; i<parameters_count; i++)
     {
      //---        
      if(passes_count==1)
        {
         string current_value="";      //   
         static int c=0,v=0,trigger=0; //   
         //---  ,     
         if(StringFind(parameters_list[i],"CriterionSelectionRule",0)>=0)
           {
            include_criteria_list=true;
            continue;
           }
         //---      ,
         //       AND ()
         if(CriterionSelectionRule==RULE_AND && i==parameters_count-1)
            CalculateUsedCriteria();
         //---       
         if(include_criteria_list)
           {
            //---   
            if(trigger==0)
              {
               equality_sign_index=StringFind(parameters_list[i],"=",0)+1;          //    '='  
               current_value =StringSubstr(parameters_list[i],equality_sign_index); //   
               //---
               criteria[c]=(int)StringToInteger(current_value);
               trigger=1; //    
               c++;
               continue;
              }
            //---   
            if(trigger==1)
              {
               equality_sign_index=StringFind(parameters_list[i],"=",0)+1;          //    '='  
               current_value=StringSubstr(parameters_list[i],equality_sign_index);  //   
               //---           
               criteria_values[v]=StringToDouble(current_value);
               trigger=0; //    
               v++;
               continue;
              }
           }
        }
      //---     
      if(ParameterEnabledForOptimization(parameters_list[i]))
        {
         //---    
         optimized_parameters_count++;
         //---      
         //        ( )
         if(passes_count==1)
           {
            //---     
            ArrayResize(parameter_names,optimized_parameters_count);
            //---    '='          
            equality_sign_index=StringFind(parameters_list[i],"=",0);
            //---   
            parameter_names[i]=StringSubstr(parameters_list[i],0,equality_sign_index);
           }
         //---     
         ArrayResize(parameter_values,optimized_parameters_count);
         //---    '='
         equality_sign_index=StringFind(parameters_list[i],"=",0)+1;
         //---   
         parameter_values[i]=StringSubstr(parameters_list[i],equality_sign_index);
        }
     }
//---      
   for(int i=0; i<STAT_VALUES_COUNT; i++)
      StringAdd(string_to_write,DoubleToString(stat_values[i],2)+",");
//---      
   for(int i=0; i<optimized_parameters_count; i++)
     {
      //---     ,   
      if(i==optimized_parameters_count-1)
        {
         StringAdd(string_to_write,parameter_values[i]);
         break;
        }
      //---   
      else
         StringAdd(string_to_write,parameter_values[i]+",");
     }
//---     -   
   if(passes_count==1)
      WriteOptimizationReport(parameter_names);
//---      
   WriteOptimizationResults(string_to_write);
  }
//+------------------------------------------------------------------+
//|                               |
//+------------------------------------------------------------------+
void WriteOptimizationReport(string &parameter_names[])
  {
   int files_count=1; //   
//---     
   string headers="#,PROFIT,TOTAL DEALS,PROFIT FACTOR,EXPECTED PAYOFF,EQUITY DD MAX REL%,RECOVERY FACTOR,SHARPE RATIO,";
//---    
   for(int i=0; i<ArraySize(parameter_names); i++)
     {
      if(i==ArraySize(parameter_names)-1)
         StringAdd(headers,parameter_names[i]);
      else
         StringAdd(headers,parameter_names[i]+",");
     }
//---        -    
   OptimizationResultsPath=CreateOptimizationResultsFolder(files_count);
//---     , 
   if(OptimizationResultsPath=="")
     {
      Print("Empty path: ",OptimizationResultsPath);
      return;
     }
   else
     {
      OptimizationFileHandle=FileOpen(OptimizationResultsPath+"\optimization_results"+IntegerToString(files_count)+".csv",
                                      FILE_CSV|FILE_READ|FILE_WRITE|FILE_ANSI|FILE_COMMON,",");
      //---
      if(OptimizationFileHandle!=INVALID_HANDLE)
         FileWrite(OptimizationFileHandle,headers);
     }
  }
//+------------------------------------------------------------------+
//|                 |
//+------------------------------------------------------------------+
void WriteOptimizationResults(string string_to_write)
  {
   bool condition=false; //   
//---      
   if(CriterionSelectionRule==RULE_OR)
      condition=AccessCriterionOR();
//---    
   if(CriterionSelectionRule==RULE_AND)
      condition=AccessCriterionAND();
//---     
   if(condition)
     {
      //---       
      if(OptimizationFileHandle!=INVALID_HANDLE)
        {
         int strings_count=0; //  
         //---          
         strings_count=GetStringsCount();
         //---    
         FileWrite(OptimizationFileHandle,IntegerToString(strings_count),string_to_write);
        }
      else
         Print("      !");
     }
  }
//+------------------------------------------------------------------+
//|                             |
//+------------------------------------------------------------------+
string CreateOptimizationResultsFolder(int &files_count)
  {
   long   search_handle       =INVALID_HANDLE;        //  
   string returned_filename   ="";                    //    (/)
   string path                ="";                    //    /
   string search_filter       ="*";                   //   (* -   /)
   string root_folder         ="OPTIMIZATION_DATA\\"; //  
   string expert_folder       =EXPERT_NAME+"\\";      //  
   bool   root_folder_exists  =false;                 //    
   bool   expert_folder_exists=false;                 //    
//---    OPTIMIZATION_DATA    
   path=search_filter;
//---          \Files
   search_handle=FileFindFirst(path,returned_filename,FILE_COMMON);
//---        
   Print("TERMINAL_COMMONDATA_PATH: ",COMMONDATA_PATH);
//---    ,  
   if(returned_filename==root_folder)
     {
      root_folder_exists=true;
      Print("  "+root_folder+" .");
     }
//---    
   if(search_handle!=INVALID_HANDLE)
     {
      //---      
      if(!root_folder_exists)
        {
         //---        
         while(FileFindNext(search_handle,returned_filename))
           {
            //---  ,   
            if(returned_filename==root_folder)
              {
               root_folder_exists=true;
               Print("  "+root_folder+" .");
               break;
              }
           }
        }
      //---     
      FileFindClose(search_handle);
     }
   else
     {
      Print("     "
            "  "+COMMONDATA_PATH+" : ",ErrorDescription(GetLastError()));
     }
//---      OPTIMIZATION_DATA
   path=root_folder+search_filter;
//---      ..\Files\OPTIMIZATION_DATA\
   search_handle=FileFindFirst(path,returned_filename,FILE_COMMON);
//---       
   if(returned_filename==expert_folder)
     {
      expert_folder_exists=true; //  
      Print("  "+expert_folder+" .");
     }
//---    
   if(search_handle!=INVALID_HANDLE)
     {
      //---       
      if(!expert_folder_exists)
        {
         //---      DATA_OPTIMIZATION     
         while(FileFindNext(search_handle,returned_filename))
           {
            //---  ,   
            if(returned_filename==expert_folder)
              {
               expert_folder_exists=true;
               Print("  "+expert_folder+" .");
               break;
              }
           }
        }
      //---     
      FileFindClose(search_handle);
     }
   else
      Print("       "+path+" .");
//---     
   path=root_folder+expert_folder+search_filter;
//---        ..\Files\OPTIMIZATION_DATA\
   search_handle=FileFindFirst(path,returned_filename,FILE_COMMON);
//---    ,  
   if(StringFind(returned_filename,"optimization_results",0)>=0)
      files_count++;
//---    
   if(search_handle!=INVALID_HANDLE)
     {
      //---      
      while(FileFindNext(search_handle,returned_filename))
         files_count++;
      //---
      Print(" : ",files_count);
      //---     
      FileFindClose(search_handle);
     }
   else
      Print("       "+path+" ");
//---      
//        OPTIMIZATION_DATA
   if(!root_folder_exists)
     {
      if(FolderCreate("OPTIMIZATION_DATA",FILE_COMMON))
        {
         root_folder_exists=true;
         Print("   ..\Files\OPTIMIZATION_DATA\\");
        }
      else
        {
         Print("     OPTIMIZATION_DATA: ",
               ErrorDescription(GetLastError()));
         return("");
        }
     }
//---    
   if(!expert_folder_exists)
     {
      if(FolderCreate(root_folder+EXPERT_NAME,FILE_COMMON))
        {
         expert_folder_exists=true;
         Print("   ..\Files\OPTIMIZATION_DATA\\"+expert_folder);
        }
      else
        {
         Print("     ..\Files\\"+expert_folder+"\: ",
               ErrorDescription(GetLastError()));
         return("");
        }
     }
//---    
   if(root_folder_exists && expert_folder_exists)
     {
      //---  ,         
      return(root_folder+EXPERT_NAME);
     }
//---
   return("");
  }
//+------------------------------------------------------------------+
//|                                      |
//+------------------------------------------------------------------+
int GetStringsCount()
  {
   int   strings_count =0;  //  
   ulong offset        =0;  //      
//---    
   FileSeek(OptimizationFileHandle,0,SEEK_SET);
//---           
   while(!FileIsEnding(OptimizationFileHandle) || !IsStopped())
     {
      //---   
      while(!FileIsLineEnding(OptimizationFileHandle) || !IsStopped())
        {
         //---  
         FileReadString(OptimizationFileHandle);
         //---   
         offset=FileTell(OptimizationFileHandle);
         //---    
         if(FileIsLineEnding(OptimizationFileHandle))
           {
            //---    , 
            //        ,     
            if(!FileIsEnding(OptimizationFileHandle))
               offset++;
            //---  
            FileSeek(OptimizationFileHandle,offset,SEEK_SET);
            //---   
            strings_count++;
            break;
           }
        }
      //---    ,    
      if(FileIsEnding(OptimizationFileHandle))
         break;
     }
//---       
   FileSeek(OptimizationFileHandle,0,SEEK_END);
//---  - 
   return(strings_count);
  }
//+------------------------------------------------------------------+
//|                       |
//+------------------------------------------------------------------+
void CalculateUsedCriteria()
  {
   UsedCriteriaCount=0; // 
//---      
   for(int i=0; i<ArraySize(criteria); i++)
     {
      //---  
      if(criteria[i]!=C_NO_CRITERION)
         UsedCriteriaCount++;
     }
  }
//+------------------------------------------------------------------+
//|                 |
//+------------------------------------------------------------------+
bool AccessCriterionOR()
  {
//---        , 
//            
   for(int i=0; i<ArraySize(criteria); i++)
     {
     //---
      if(criteria[i]==C_NO_CRITERION)
         continue;
      //--- PROFIT
      if(criteria[i]==C_STAT_PROFIT)
        {
         if(stat_values[0]>criteria_values[i])
            return(true);
        }
      //--- TOTAL DEALS
      if(criteria[i]==C_STAT_DEALS)
        {
         if(stat_values[1]>criteria_values[i])
            return(true);
        }
      //--- PROFIT FACTOR
      if(criteria[i]==C_STAT_PROFIT_FACTOR)
        {
         if(stat_values[2]>criteria_values[i])
            return(true);
        }
      //--- EXPECTED PAYOFF
      if(criteria[i]==C_STAT_EXPECTED_PAYOFF)
        {
         if(stat_values[3]>criteria_values[i])
            return(true);
        }
      //--- EQUITY DD REL PERC
      if(criteria[i]==C_STAT_EQUITY_DDREL_PERCENT)
        {
         if(stat_values[4]<criteria_values[i])
            return(true);
        }
      //--- RECOVERY FACTOR
      if(criteria[i]==C_STAT_RECOVERY_FACTOR)
        {
         if(stat_values[5]>criteria_values[i])
            return(true);
        }
      //--- SHARPE RATIO
      if(criteria[i]==C_STAT_SHARPE_RATIO)
        {
         if(stat_values[6]>criteria_values[i])
            return(true);
        }
     }
//---     
   return(false);
  }
//+------------------------------------------------------------------+
//|                           |
//+------------------------------------------------------------------+
bool AccessCriterionAND()
  {
   int count=0; //  
//---        ,
//            
   for(int i=0; i<ArraySize(criteria); i++)
     {
      //---    ,    
      if(criteria[i]==C_NO_CRITERION)
         continue;
      //--- PROFIT
      if(criteria[i]==C_STAT_PROFIT)
        {
         if(stat_values[0]>criteria_values[i])
           {
            count++;
            if(count==UsedCriteriaCount)
               return(true);
           }
        }
      //--- TOTAL DEALS
      if(criteria[i]==C_STAT_DEALS)
        {
         if(stat_values[1]>criteria_values[i])
           {
            count++;
            if(count==UsedCriteriaCount)
               return(true);
           }
        }
      //--- PROFIT FACTOR
      if(criteria[i]==C_STAT_PROFIT_FACTOR)
        {
         if(stat_values[2]>criteria_values[i])
           {
            count++;
            if(count==UsedCriteriaCount)
               return(true);
           }
        }
      //--- EXPECTED PAYOFF
      if(criteria[i]==C_STAT_EXPECTED_PAYOFF)
        {
         if(stat_values[3]>criteria_values[i])
           {
            count++;
            if(count==UsedCriteriaCount)
               return(true);
           }
        }
      //--- EQUITY DD REL PERC
      if(criteria[i]==C_STAT_EQUITY_DDREL_PERCENT)
        {
         if(stat_values[4]<criteria_values[i])
           {
            count++;
            if(count==UsedCriteriaCount)
               return(true);
           }
        }
      //--- RECOVERY FACTOR
      if(criteria[i]==C_STAT_RECOVERY_FACTOR)
        {
         if(stat_values[5]>criteria_values[i])
           {
            count++;
            if(count==UsedCriteriaCount)
               return(true);
           }
        }
      //--- SHARPE RATIO
      if(criteria[i]==C_STAT_SHARPE_RATIO)
        {
         if(stat_values[6]>criteria_values[i])
           {
            count++;
            if(count==UsedCriteriaCount)
               return(true);
           }
        }
     }
//---     
   return(false);
  }
//+------------------------------------------------------------------+
//| ,                  |
//+------------------------------------------------------------------+
bool ParameterEnabledForOptimization(string parameter_string)
  {
   bool enable;
   long value,start,step,stop;
//---    '='  
   int equality_sign_index=StringFind(parameter_string,"=",0);
//---   
   ParameterGetRange(StringSubstr(parameter_string,0,equality_sign_index),
                     enable,value,start,step,stop);
//---   
   return(enable);
  }
//+------------------------------------------------------------------+
